/* SDCal, version 01
 * Copyright (C) 2022 Arktis Desktop Project, GNU GPL v2 or any later.*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <langinfo.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/xpm.h>
#include <X11/Xft/Xft.h>
#include "SDCal.h"

#define SATURDAY 6
#define FIRST_MISSING_DAY 639799
#define NUMBER_MISSING_DAYS 11
#define is_leap_year(y) \
    ((y) <= 1752 ? !((y) % 4) : \
    (!((y) % 4) && ((y) % 100)) || !((y) % 400))
#define centuries_since_1700(y) \
    ((y) > 1700 ? (y) / 100 - 17 : 0)
#define quad_centuries_since_1700(y) \
    ((y) > 1600 ? ((y) - 1600) / 400 : 0)
#define leap_years_since_year_1(y) \
    ((y) / 4 - centuries_since_1700(y) + quad_centuries_since_1700(y))

static Display *display;
static Window root, window;
static GC gc;
static Pixel bk_px, mk_px, mb_px;
static Pixmap back_pm, mask_pm, all_pm, subb_pm, day_pm;
static Pixel GetColor(char *);
static XftColor GetXftC(char *);
static XftDraw *caldraw;
static XftFont *font, *hfont, *ffont, *lfont;
static int ix, iy, eu = 0;
static int days[] = { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };

int get_x(void) {
	FILE *fp;
	int x = 200;
	if ((fp = fopen("/root/.SDCalendar/conf/dx", "r")) == NULL)
		fprintf(stderr, "SDCal: conf x missing\n");
	if (fscanf(fp, "%d\n", &x) == -1)
		fprintf(stderr, "SDCal: conf error\n");
	fclose(fp);
	return x;
}

int get_y(void) {
	FILE *fp;
	int y = 200;
	if ((fp = fopen("/root/.SDCalendar/conf/dy", "r")) == NULL)
		fprintf(stderr, "SDCal: conf y missing\n");
	if (fscanf(fp, "%d\n", &y) == -1)
		fprintf(stderr, "SDCal: conf error\n");
	fclose(fp);
	return y;
}

void CreateWindow(int ac, char *av[]) {
	int zero = 0;
	int screen;
	XSizeHints size_hints;
	XGCValues gcv;
	unsigned long gcm;
	int result;
	display = XOpenDisplay(NULL);
	if (display == NULL) {
		fprintf(stderr, "SDCal: could not open display ``%s'' Exit.\n",
				XDisplayName(NULL));
		exit(EXIT_FAILURE);
	}
	screen = DefaultScreen(display);
	root = RootWindow(display, screen);
	bk_px = GetColor("black");
	mk_px = GetColor(option.mc);
	mb_px = GetColor(option.mb);
	XpmAttributes attributes;
	attributes.valuemask = XpmSize;
	if (option.bg_pix[0] != '\0') {
		char db[128];
		sprintf(db, "/root/.SDCalendar/data/%s", option.bg_pix);
		result = XpmReadFileToPixmap(display, root, db, &back_pm, &mask_pm, 
										&attributes);
		if (result != XpmSuccess) {
			fprintf(stderr, "SDCal: pixmap not found in data folder\n");
			exit(EXIT_FAILURE);
		} else {
			ix = attributes.width;
			iy = attributes.height;
		}
	} else {
		fprintf(stderr, "SDCal: background undefined.. exit\n");
		exit(EXIT_FAILURE);
	}
	size_hints.flags = PMinSize | PMaxSize | PPosition;
	size_hints.min_width = size_hints.max_width = ix;
	size_hints.min_height = size_hints.max_height = iy;
	XpmFreeAttributes(&attributes);
	all_pm = XCreatePixmap(display, root, ix, iy, DefaultDepth(display, screen));
	XWMGeometry(display, screen, "", NULL, 0, &size_hints, &zero, &zero,
				&size_hints.width, &size_hints.height, &zero); 
	window = XCreateSimpleWindow(display, root, get_x(), get_y(),
			size_hints.width, size_hints.height, 0, bk_px, bk_px);
	XSetWMNormalHints(display, window, &size_hints);
	gcm = GCForeground | GCBackground;
	gcv.foreground = mk_px;
	gcv.background = bk_px;
	gc = XCreateGC(display, root, gcm, &gcv);
	XSetCommand(display, window, av, ac);
	XCopyArea(display, back_pm, all_pm, gc, 0, 0, ix, iy, 0, 0);
}

void ShowCal(void) {
	char wcommand[32];
	sprintf(wcommand, "sdcalx %d %d %d %d", get_x(), get_y(), ix, iy);
	pid_t pid = fork();
	switch (pid) {
		case 0:
			execlp ("/bin/sh", "sh", "-c", wcommand, NULL);
			fprintf (stderr, "SDCal: Exec failed.\n");
			exit (0);
		break;
		case -1:
			fprintf (stderr, "SDCal: Fork failed.\n");
		break;
	}
}

int d_in_y(int d, int m, int y) {
	int i;
	for (i = 1; i < m; i++) {
		if (i == 2)
			d += (is_leap_year(y)) ? 29 : 28;
		else
			d += days[i];
	}
	return d;
}

int d_in_w(int d, int m, int y) {
	int total_days;
	total_days = (int)(y - 1) * 365 + leap_years_since_year_1(y - 1)
					+ d_in_y(d, m, y);
	return (((total_days - 1 + SATURDAY - eu) - NUMBER_MISSING_DAYS) % 7);
}

void PaintCal(void) {
	int screen = DefaultScreen(display);
	char *calfontname = option.fn;
	font = XftFontOpenName(display, screen, calfontname);
	if (font == NULL) {
		fprintf(stderr, "SDCal: trouble loading font\n");
	}
	if (!option.hh) {
		char *headerfontname = option.fh;
		hfont = XftFontOpenName(display, screen, headerfontname);
		if (hfont == NULL) {
			fprintf(stderr, "SDCal: trouble loading header font\n");
		}
	} else {
		hfont = font;
	}
	if (!option.fy) {
		char *footerfontname = option.ff;
		ffont = XftFontOpenName(display, screen, footerfontname);
		if (ffont == NULL) {
			fprintf(stderr, "SDCal: trouble loading footer font\n");
		}
	} else {
		ffont = font;
	}
	if (!option.dd) {
		char *dlinefont = option.fd;
		lfont = XftFontOpenName(display, screen, dlinefont);
		if (lfont == NULL) {
			fprintf(stderr, "SDCal: trouble loading dayline font\n");
		}
	} else {
		lfont = font;
	}
	Colormap colormap = DefaultColormap(display, screen);
	Visual *vis;
	vis = DefaultVisual(display, screen);
	caldraw = XftDrawCreate(display, all_pm, vis, colormap);
	XGlyphInfo extents;
	XftTextExtentsUtf8(display, font, (const FcChar8 *) "0", strlen("0"), 
						&extents);
	int glyph_w, glyph_h, sct_w;
	glyph_w = extents.xOff;
	glyph_h = extents.height; 
	sct_w = (glyph_w * 2) + option.dx;/* .dx default is 7 */
	int i = 0, j = 0;
	if (!option.ei) {  /* ? */
		if (option.ee == 0)
			eu = 0;
		else
			eu = 1;
	} else {
		int fwd = (*nl_langinfo(_NL_TIME_FIRST_WEEKDAY) - 1);
		if (fwd == 1)
			eu = 1;
		else
			eu = 0;
	}
	struct tm *time_struct;
	time_t curtime;
	curtime = time(NULL);
	time_struct = localtime(&curtime);
	int m, y, tsday;
	y = time_struct->tm_year + 1900;
	m = time_struct->tm_mon + 1;
	tsday = time_struct->tm_mday;
	char month[STRING_M];  /* ? */
	int mo = option.pm;
	char *nth;
	if (option.pm !=0)
		nth = nl_langinfo(MON_1 + mo - 1);
	if (option.pm !=0 && option.py !=0) 
		sprintf(month, "%s", nth);
	else
		strftime (month, sizeof(month), option.hl, time_struct);
	char year[STRING_M];
	if (option.pm !=0 && option.py !=0) { 
		sprintf(year, "%d", option.py);
		y = option.py;
		m = option.pm;
		}
	else
		strftime (year, sizeof(year), option.rh, time_struct);
	int hb;
	if (option.cl == 1) {
		hb = option.hx;
	} else {
		XGlyphInfo extd;
		XftTextExtentsUtf8(display, hfont, (const FcChar8 *) month, 
							strlen(month), &extd);
		int hdr_w = extd.xOff;
		if (ix > hdr_w && option.cl == 2)
			hb = (ix - hdr_w) / 2;
		else if (ix > hdr_w && option.cl == 3)
			hb = ix - hdr_w - option.hx;
		else
			hb = option.hx;
	}
	char dummy[] = " ";
	if (!strstr(dummy, month)) {
		if (!option.cd) {
			XftColor c3 = GetXftC(option.cb);
			XftDrawStringUtf8(caldraw, &c3, hfont, hb + 1, option.hy + 1,
							(const XftChar8 *) month, strlen(month));
		}
		XftColor c3 = GetXftC(option.ca);
		XftDrawStringUtf8(caldraw, &c3, hfont, hb, option.hy,
						(const XftChar8 *) month, strlen(month));
	}

	if (!option.sh) {
		XftColor c3 = GetXftC(option.hs);
		XftDrawStringUtf8(caldraw, &c3, ffont, option.rx + 1, option.ry + 1,
							(const XftChar8 *) year, strlen(year));
	}
	XftColor c3 = GetXftC(option.hc);
	XftDrawStringUtf8(caldraw, &c3, ffont, option.rx, option.ry,
						(const XftChar8 *) year, strlen(year));
	int e, day;
	char *fgc, *bgc;
	char abb[STRING_M];   /* wide ? */
	for (e = 0; e < 7; e++) {
		if (eu)
			day = (e + 1) % 7;
		else
			day = e;
		sprintf(abb, "%s", nl_langinfo(ABDAY_1 + day));
		XGlyphInfo extb;
		XftTextExtentsUtf8(display, lfont, (const FcChar8 *) abb, strlen(abb), 
						&extb);
		int lbl_w, padx, sct;
		lbl_w = extb.xOff;
		sct = sct_w * e;
		padx = (sct_w - lbl_w) / 2;
		if (!option.aa) {
			bgc = (option.ab) ? option.sc : option.as;
			XftColor c3 = GetXftC(bgc);
			XftDrawStringUtf8(caldraw, &c3, lfont, option.xc + 1 + 
				sct + padx, option.yc + 1, (const XftChar8 *) abb, strlen(abb));
		}
		fgc = (option.ab) ? option.tc : option.ac;
		XftColor c3 = GetXftC(fgc);
		XftDrawStringUtf8(caldraw, &c3, lfont, option.xc + sct + 
			padx, option.yc, (const XftChar8 *) abb, strlen(abb));
	}
	int dw = d_in_w(1, m, y);
	char buff[3];
	int soc, td, mxd, px;
	if (m == 2)
		mxd = (is_leap_year(y)) ? 29 : 28;
	else
		mxd = days[m];
	char qq =' ';
	td = 1;
	for (i = 0; i < 6; i++) {
		if (i == 0)
			soc = dw;  /* first day */
		else 
			soc = 0;
		for (j = soc; j < 7; j++) {
			if (td > mxd)
				sprintf(buff, "%c", qq); /* done, fill with spaces.. */
			else {
				sprintf(buff, "%d", td);
				px = sct_w * j;
				XGlyphInfo extc;
				XftTextExtentsUtf8(display, font, (const FcChar8 *) buff,
						strlen(buff), &extc);
				int dbl_w, padd, sct_h, nrl, lne, ytp, mpw, mph;
				dbl_w = extc.xOff;
				padd = (sct_w - dbl_w) / 2;
				sct_h = glyph_h + option.dy;
				nrl = option.yc + glyph_h + option.dl;
				lne = i * sct_h;
				ytp = (i * sct_h) + (option.dy / 2) - sct_h;
				mpw = (glyph_w * 2) + option.mp;
				mph = glyph_h + option.mp;
				char cs;
				if ((option.ar == 4) || (option.ar == 5))
					cs = JoinBevel;
				else
					cs = JoinMiter;
				if ((td == tsday) && (option.pm == 0 && option.py == 0)) {
					if (option.ar == 6) {
						int dpx;
						XpmAttributes attributes;
						attributes.valuemask = XpmSize;
						if (option.day_pix[0] != '\0') {
							char tb[128];
							sprintf(tb, "/root/.SDCalendar/data/%s", option.day_pix);
							dpx = XpmReadFileToPixmap(display, root, tb, 
											&day_pm, &mask_pm, &attributes);
							if (dpx != XpmSuccess) {
								fprintf(stderr, "SDCal: marker error\n");
								exit(EXIT_FAILURE);
							} else {
								int mhgt, mwdt, mkx, mky;
								mwdt = attributes.width;
								mhgt = attributes.height;
								mkx = option.xc + px - option.mp;
								mky = nrl + ytp - option.mp;
								XCopyArea(display, day_pm, all_pm, gc, 0, 0, 
											mwdt, mhgt, mkx, mky);
							}
						}
						XpmFreeAttributes(&attributes);
					} else if ((option.ar == 0) || (option.ar == 1)) {
						if (option.ar == 1) {
							XSetForeground(display, gc, mb_px);
							XFillArc(display, all_pm, gc, option.xc + px + 
									((sct_w - mpw) / 2), 
									nrl + ytp + ((sct_h - mpw) / 2), 
									mpw, mpw, 0, 360 * 64);
						}
						XSetForeground(display, gc, mk_px);
						XSetLineAttributes(display, gc, option.mw, LineSolid, 
										CapRound, JoinRound);
						XDrawArc(display, all_pm, gc, option.xc + px + 
									((sct_w - mpw) / 2), 
									nrl + ytp + ((sct_h - mpw) / 2), 
									mpw, mpw, 0, 360 * 64);
					} else {
						if ((option.ar == 3) || (option.ar == 5)) {
							XSetForeground(display, gc, mb_px);
							XFillRectangle(display, all_pm, gc, option.xc +
									px + ((sct_w - mpw) / 2),
									nrl + ytp + ((sct_h - mph) / 2), mpw, mph);
						}
						XSetLineAttributes(display, gc, option.mw, LineSolid, 
										CapButt, cs);
						XSetForeground(display, gc, mk_px);
						XDrawRectangle(display, all_pm, gc, option.xc + 
									px + ((sct_w - mpw) / 2), 
									nrl + ytp + ((sct_h - mph) / 2), mpw, mph);
					}
				}
				if ((td == tsday) && (option.pm == 0 && option.py == 0)
					&& (!option.ii)) {
					if (!option.ss) {
						XftColor c3 = GetXftC(option.is);
						XftDrawStringUtf8(caldraw, &c3, font, option.xc + px +
									padd + 1, nrl + lne + 1,
									(const XftChar8 *) buff, strlen(buff));
					}
					XftColor c3 = GetXftC(option.ic);
					XftDrawStringUtf8(caldraw, &c3, font, option.xc + px + padd,
						nrl + lne, (const XftChar8 *) buff, strlen(buff));
				} else {
					if (!option.ss) {
						if (((eu == 1) && (j == 6)) || ((eu == 0) && (j == 0))) 
							bgc = option.wb;
						else if (((eu == 1) && (j == 5)) || 
									((eu == 0) && (j == 6)))
							bgc = option.ub;
						else 
							bgc = option.sc;
						XftColor c3 = GetXftC(bgc);
						XftDrawStringUtf8(caldraw, &c3, font, option.xc + 
								px + 1 + padd, nrl + lne + 1, 
								(const XftChar8 *) buff, strlen(buff));
					}
					if (((eu == 1) && (j == 6)) || ((eu == 0) && (j == 0)))
						fgc = option.wa;
					else if (((eu == 1) && (j == 5)) || 
								((eu == 0) && (j == 6)))
						fgc = option.ua;
					else 
						fgc = option.tc;
					XftColor c3 = GetXftC(fgc);
					XftDrawStringUtf8(caldraw, &c3, font, option.xc + px + padd,
						nrl + lne, (const XftChar8 *) buff, strlen(buff));
				}
				if (td <= mxd)
					td++;
			}
		}
	}
	/* export the pixmap and clean up */
	subb_pm = XCreatePixmap(display, root, ix, iy, DefaultDepth(display, screen));
	XCopyArea(display, all_pm, subb_pm, gc, 0, 0, ix, iy, 0, 0);
	char *filenm;
	if (option.pix)
		filenm = "/root/.SDCalendar/tmp/bg.xpm"; 
	else
		filenm = "/root/.SDCalendar/tmp/today.xpm"; 
	XpmAttributes attributes; 
	attributes.valuemask = XpmSize;
	attributes.width = ix;
	attributes.height = iy;
	XpmWriteFileFromPixmap(display, filenm, subb_pm, mask_pm, &attributes);
	XpmFreeAttributes(&attributes);
	XftColorFree (display, vis, colormap, &c3);
	if (!option.hh)
		XftFontClose(display, hfont);
	if (!option.fy)
		XftFontClose(display, ffont);
	if (!option.dd)
		XftFontClose(display, lfont);
	XftFontClose(display, font);
	XftDrawDestroy(caldraw);
	XFreeGC(display, gc);
	XFreePixmap(display, all_pm);
	XFreePixmap(display, back_pm);
	XFreePixmap(display, subb_pm);
	if (option.ar == 6) {
		XFreePixmap(display, day_pm);
	}
	XDestroyWindow(display, window);
	XCloseDisplay(display);
}


static Pixel GetColor(char *name) {
	XColor color;
	XWindowAttributes attributes;
	XGetWindowAttributes(display, root, &attributes);
	color.pixel = 0;
	if (!XParseColor(display, attributes.colormap, name, &color)) {
		fprintf(stderr, "SDCal: could not parse color ``%s''.\n", name);
	} else if (!XAllocColor(display, attributes.colormap, &color)) {
		fprintf(stderr, "SDCal: could not allocate ``%s''.\n", name);
	}
	return color.pixel;
}

/* fixme.. */
static XftColor GetXftC(char *xname) {
	int screen = DefaultScreen(display);
	XColor c;
	XftColor c3;
	c.pixel = GetColor(xname);
	XQueryColor(display, DefaultColormap(display, screen), &c);
	c3.pixel = c.pixel;
	c3.color.red = c.red;
	c3.color.green = c.green;
	c3.color.blue = c.blue;
	c3.color.alpha = 0xffff;
	return c3;
}

